﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.IO.Ports;

namespace AT89C2051Programmer
{
    public partial class Form1 : Form
    {
        SerialPortForm sForm = null;
        int mode = 0;
        int address = 0;
        byte[] chipMem = new byte[2048];
        private string loadFileName = "";

        public Form1()
        {
            InitializeComponent();
        }

        private void Form1_Load(object sender, EventArgs e)
        {
            for (int i = 0; i < 2048; i++)
                chipMem[i] = 0xFF;

            loadFileName = Properties.Settings.Default.loadFileName;
            comPortComboBox.Items.AddRange(SerialPort.GetPortNames());
            comPortComboBox.SelectedIndex = comPortComboBox.FindStringExact(Properties.Settings.Default.comPort);
            sForm = new SerialPortForm(this);
            sForm.setPort(Properties.Settings.Default.comPort);
            // sForm.Show();
        }

        private void Form1_Resize(object sender, EventArgs e)
        {
            if (this.Height > 186)
            {
                fileContentsText.Height = this.Height - 136;
                fileContentsText.Visible = true;
            }
            else
            {
                fileContentsText.Visible = false;
            }

        }

        private void comPortComboBox_MouseEnter(object sender, EventArgs e)
        {
            comPortComboBox.Items.Clear();
            comPortComboBox.Items.AddRange(SerialPort.GetPortNames());
            comPortComboBox.SelectedIndex = comPortComboBox.FindStringExact(Properties.Settings.Default.comPort);
        }

        private void comPortComboBox_SelectedIndexChanged(object sender, EventArgs e)
        {
            Properties.Settings.Default.comPort = comPortComboBox.Text;
            Properties.Settings.Default.Save();
            if (sForm != null)
            {
                sForm.setPort(Properties.Settings.Default.comPort);
            }
        }

        private void addText(string s)
        {
            Utils.SetControlPropertyValue(fileContentsText, "Text", fileContentsText.Text + s);
        }

        private void displayMem(int length)
        {
            if (length > 2048) length = 2048;

            string s = "0000: ";

            int a = 0;
            while (a < length)
            {
                if (a > 0 && a % 16 == 0)
                {
                    s += "\r\n" + a.ToString("X4") + ": ";
                }
                s += chipMem[a++].ToString("X2") + " ";
            }
            Utils.SetControlPropertyValue(fileContentsText, "Text", s);
        }

        private void serialCallback(bool success, byte[] buffer)
        {
            string s;
            int i;

            if (success)
            {
                switch (buffer[0])
                {
                    case 0x11: // reading signature bytes
                        s = "Chip signature bytes: ";
                        for (i = 1; i < buffer.Length; i++)
                        {
                            s += buffer[i].ToString("X2") + " ";
                        }
                        Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                        break;
                    case 0x12: // reading chip data
                        address = 256 * (int)(buffer[1]) + (int)(buffer[2]);
                        int length = buffer[3];
                        Utils.SetControlPropertyValue(progressBar1, "Value", address + length > 2048 ? 2048 : address + length);


                        for (i = 0; 4 + i < buffer.Length && address + i < 2048; i++)
                        {
                            if (mode == 1) // reading chip
                            {
                                chipMem[address + i] = buffer[4 + i];
                            }
                            else if (mode == 3) // verifying chip
                            {
                                if (buffer[4 + i] != chipMem[address + i])
                                {
                                    s = "Verify error at address " + (address + i).ToString("X4");
                                    Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                                    mode = 0;
                                    sendCommand(0x13); // power down
                                }
                            }
                        }

                        if (buffer.Length != length + 4)
                        {
                            displayMem(buffer.Length > length + 4 ? address + buffer.Length - 4 : address + length);
                            addText("\r\nBad packet length in chip data response");
                            mode = 0;
                            sendCommand(0x13); // power down
                        }
                        else
                        {
                            if (address + length < 2048)
                            {
                                sendCommand(0x12); // keep reading next chunk till whole 2K read
                            }
                            else
                            {
                                Utils.SetControlPropertyValue(progressBar1, "Value", 2048);
                                int usedMem = 2048;
                                while (usedMem > 0 && chipMem[usedMem - 1] == 0xFF)
                                    usedMem--;

                                if (usedMem == 0)
                                {
                                    s = "Memory appears to be blank";
                                    Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                                }
                                else
                                {
                                    if (mode == 3) // finished verifying
                                    {
                                        s = "Verified okay";
                                        Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                                    }
                                    else
                                    {
                                        displayMem(usedMem);
                                    }
                                }
                                mode = 0;
                                Utils.SetControlPropertyValue(progressLabel, "Visible", false);
                                Utils.SetControlPropertyValue(progressBar1, "Value", 0);
                                sendCommand(0x13); // power down
                            }
                        }

                        break;
                    case 0x13: // responding to power down command
                        mode = 0;
                        break;
                    case 0x14: // responding to power up command
                        if (mode == 1 || mode == 3) // reading or verifying chip data
                        {
                            address = 0;
                            sendCommand(0x12); // read first 16 bytes of data
                        }
                        else if (mode == 2) // writing chip data
                        {
                            address = 0;
                            writeToChip();
                        }
                        break;
                    case 0x16: // responding to erase chip command
                        s = "Erase command sent";
                        Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                        sendCommand(0x13); // power down
                        break;
                    case 0x17: // responding to write lock bit 1 command
                        s = "Write lock bit 1 command sent";
                        Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                        sendCommand(0x13); // power down
                        break;
                    case 0x18: // responding to write lock bit 2 command
                        s = "Write lock bit 2 command sent";
                        Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                        sendCommand(0x13); // power down
                        break;
                    case 0x21: // responding to write data command
                        if (mode != 2)
                        {
                            s = "Unexpected write command response when not in write mode";
                            Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                            sendCommand(0x13); // power down
                        }
                        else
                        {
                            if (256 * (int)(buffer[1]) + (int)(buffer[2]) != address)
                            {
                                s = "Address out of sync in write command response";
                                Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                                sendCommand(0x13); // power down
                            }
                            else if (address >= progressBar1.Maximum) // finished writing
                            {
                                Utils.SetControlPropertyValue(progressLabel, "Visible", false);
                                Utils.SetControlPropertyValue(progressBar1, "Value", 0);
                                sendCommand(0x13); // power down
                            }
                            else
                            {
                                Utils.SetControlPropertyValue(progressBar1, "Value", address > progressBar1.Maximum ? progressBar1.Maximum : address);
                                writeToChip();
                            }
                        }
                        break;
                    case 0xFF:
                        s = "Arduino Mega reports unrecognized command: " + buffer[1].ToString("X2");
                        Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                        break;
                    default:
                        s = "Unrecognized packet received: ";
                        for (i = 1; i < buffer.Length; i++)
                        {
                            s += buffer[i].ToString("X2") + " ";
                        }
                        Utils.SetControlPropertyValue(fileContentsText, "Text", s);
                        break;
                }
            }
            else
            {
                Utils.SetControlPropertyValue(fileContentsText, "Text", "No response from Arduino Mega 2560 - check port and connections");
            }
        }

        private void sendCommand(byte command)
        {
            byte[] buffer = new byte[1];
            buffer[0] = command;
            SerialPortForm.callbackDelegate d = new SerialPortForm.callbackDelegate(serialCallback);
            sForm.send(buffer, 1000, d);
        }

        private void readSignatureButton_Click(object sender, EventArgs e)
        {
            sendCommand(0x11);
        }

        private void eraseChipButton_Click(object sender, EventArgs e)
        {
            sendCommand(0x16);
        }

        private void powerDownButton_Click(object sender, EventArgs e)
        {
            sendCommand(0x13);
        }

        private void writeLock1Button_Click(object sender, EventArgs e)
        {
            sendCommand(0x17);
        }

        private void writeLock2Button_Click(object sender, EventArgs e)
        {
            sendCommand(0x18);
        }

        private void readProgramButton_Click(object sender, EventArgs e)
        {
            progressLabel.Text = "Reading chip:";
            progressLabel.Visible = true;
            progressBar1.Maximum = 2048;
            progressBar1.Value = 0;
            mode = 1; // reading chip data
            sendCommand(0x14);
        }

        private void loadFile()
        {

            if (loadFileName.Length == 0)
            {
                MessageBox.Show("Please choose a file to load", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                fileContentsText.Text = "";
                return;
            }

            if (Path.GetExtension(loadFileName).ToLower() != ".hex")
            {
                MessageBox.Show("Please choose a hex file (extension .hex)", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                fileContentsText.Text = "";
                return;
            }

            StreamReader sr = null;
            try
            {
                sr = new StreamReader(loadFileName);
            }
            catch (FileNotFoundException)
            {
                fileContentsText.Text = "";
                return;
            }

            fileContentsText.Text = sr.ReadToEnd();
            sr.Close();

            // test validity of hex file
            int lineNumber = 0;
            int maxAddress = 0;
            foreach (string line in fileContentsText.Lines)
            {
                lineNumber++;

                if (line.Length == 0) continue;

                if (line.Length < 11)
                {
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file is too short", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                if (line[0] != ':')
                {
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file does not start with a colon", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                foreach (char c in line.Substring(1))
                {
                    if (c >= '0' && c <= '9') continue;
                    if (c >= 'A' && c <= 'F') continue;
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file contains non-hex character(s)", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                byte checksum = 0x00;
                for (int i = 1; i < line.Length; i += 2)
                {
                    checksum += Convert.ToByte(line.Substring(i, 2), 16);
                }
                if (checksum != 0x00)
                {
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file has bad checksum", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                int byteCount = Convert.ToByte(line.Substring(1, 2), 16);
                if (line.Length != 2 * byteCount + 11)
                {
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file is wrong length for byte count", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                int recordType = Convert.ToByte(line.Substring(7, 2), 16);
                if (recordType > 1)
                {
                    MessageBox.Show("Line " + lineNumber.ToString() + " of file has record type " + recordType.ToString(), "Hex file is not a I8HEX file", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }
                if (recordType == 0)
                {
                    int startAddress = Convert.ToInt32(line.Substring(3, 4), 16);
                    if (startAddress + byteCount - 1 > maxAddress)
                    {
                        maxAddress = startAddress + byteCount - 1;
                    }
                }

            }

            // file seems okay - load to chipMem array

            for (int i = 0; i < 2048; i++)
            {
                chipMem[i] = 0xFF;
            }

            foreach (string line in fileContentsText.Lines)
            {
                if (line.Length == 0) continue;

                int recordType = Convert.ToByte(line.Substring(7, 2), 16);
                if (recordType == 0)
                {
                    int byteCount = Convert.ToByte(line.Substring(1, 2), 16);
                    int startAddress = Convert.ToInt32(line.Substring(3, 4), 16);
                    for (int i = 0; i < byteCount; i++)
                    {
                        if (startAddress + i >= 2048)
                        {
                            MessageBox.Show("Hex file extends beyond memory size (2048 bytes)", "Hex file format error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                            return;
                        }
                        chipMem[startAddress + i] = Convert.ToByte(line.Substring(9 + 2 * i, 2), 16);
                    }

                }
            }
            displayMem(maxAddress);
        }

        private void loadHexButton_Click(object sender, EventArgs e)
        {
            openFileDialog1.FileName = loadFileName;
            openFileDialog1.Filter = "Hex files (*.hex)|*.hex|All files (*.*)|*.*";
            openFileDialog1.FilterIndex = 1;
            DialogResult result = openFileDialog1.ShowDialog();
            if (result == DialogResult.OK)
            {
                loadFileName = openFileDialog1.FileName;
                Properties.Settings.Default.loadFileName = loadFileName;
                Properties.Settings.Default.Save();
                loadFile();
            }

        }

        private string addChecksum(string s)
        {
            byte checksum = 0x00;
            for (int i = 1; i < s.Length; i += 2)
            {
                checksum += Convert.ToByte(s.Substring(i, 2), 16);
            }
            checksum = (byte)(~checksum);
            checksum++;

            return s + checksum.ToString("X2"); 
        }

        private void writeHexButton_Click(object sender, EventArgs e)
        {
            saveFileDialog.FileName = "";
            saveFileDialog.Filter = "Hex files (*.hex)|*.hex|All files (*.*)|*.*";
            saveFileDialog.FilterIndex = 1;
            DialogResult result = saveFileDialog.ShowDialog();
            if (result == DialogResult.OK)
            {
                if (saveFileDialog.FileName.Length == 0)
                {
                    MessageBox.Show("Please enter a valid file name for save", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                if (Path.GetExtension(saveFileDialog.FileName).ToLower() != ".hex")
                {
                    MessageBox.Show("Please choose a hex file (extension .hex)", "Error", MessageBoxButtons.OK, MessageBoxIcon.Error);
                    return;
                }

                int usedMem = 2048;
                while (usedMem > 0 && chipMem[usedMem - 1] == 0xFF)
                    usedMem--;

                StreamWriter sw = new StreamWriter(saveFileDialog.FileName, false);

                for (int a = 0; a < usedMem; a += 16) // write hex file in 16-byte chunks
                {
                    int len = a + 15 < usedMem ? 16  : usedMem - a;
                    string s = ":" + len.ToString("X2") + a.ToString("X4") + "00";
                    for (int i = a; i < a + len; i++)
                    {
                        s += chipMem[i].ToString("X2");
                    }
                    sw.WriteLine(addChecksum(s));
                }
                sw.WriteLine(addChecksum(":00000001"));

                sw.Flush();
                sw.Close();
            }
        }

        private void writeToChip()
        {
            byte[] buffer = new byte[18];
            buffer[0] = 0x21; // write to chip command
            buffer[1] = 16; // indicate that this is a 16-byte chunk of data
            for (int i = 0; i < 16; i++)
            {
                buffer[2 + i] = chipMem[address + i];
            }
            address += 16; // Arduino returns 'next address to be written' which should match this value
            SerialPortForm.callbackDelegate d = new SerialPortForm.callbackDelegate(serialCallback);
            sForm.send(buffer, 1000, d);
        }

        private void writeProgramButton_Click(object sender, EventArgs e)
        {
            if (mode != 0) return;

            int usedMem = 2048;
            while (usedMem > 0 && chipMem[usedMem - 1] == 0xFF)
                usedMem--;

            if (usedMem == 0)
            {
                MessageBox.Show("Please load a hex file or read a program from a chip first", "Nothing to write", MessageBoxButtons.OK, MessageBoxIcon.Error);
                return;
            }

            progressBar1.Maximum = usedMem;

            progressLabel.Text = "Writing chip:";
            progressLabel.Visible = true;
            progressBar1.Value = 0;
            address = 0;
            mode = 2; // writing chip data
            sendCommand(0x14);
        }

        private void verifyButton_Click(object sender, EventArgs e)
        {
            Utils.SetControlPropertyValue(fileContentsText, "Text", "");
            progressLabel.Text = "Verifying chip:";
            progressLabel.Visible = true;
            progressBar1.Maximum = 2048;
            progressBar1.Value = 0;
            mode = 3; // reading chip data
            sendCommand(0x14);
        }
    }
}
